"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.CssService = void 0;
const path_1 = __importDefault(require("path"));
const vscode_css_languageservice_1 = require("vscode-css-languageservice");
const vscode_languageserver_textdocument_1 = require("vscode-languageserver-textdocument");
const vscode_languageserver_types_1 = require("vscode-languageserver-types");
const vscode_uri_1 = require("vscode-uri");
const provider_1 = require("./provider");
function readDocRange(doc, rng) {
    const lines = doc.getText().split('\n');
    return lines[rng.start.line].slice(rng.start.character, rng.end.character);
}
function findPseudoStateStart(line, lookFrom) {
    let i = lookFrom - 1;
    let res = -1;
    let openParens = 0;
    while (i !== -1) {
        if (line[i] === ':' && line[i - 1] !== ':') {
            res = i;
        }
        if (line[i] === '(') {
            openParens++;
        }
        if (line[i] === ')') {
            openParens--;
        }
        i--;
    }
    return {
        index: res,
        openParens,
    };
}
/**
 * the API for "normal" css language features fallback
 */
class CssService {
    constructor(fs) {
        this.fs = fs;
        this.inner = vscode_css_languageservice_1.getCSSLanguageService({ fileSystemProvider: this.fs.promises }); // TODO: FIX TYPE
    }
    getCompletions(document, position) {
        const cssCompsRaw = this.inner.doComplete(document, position, this.inner.parseStylesheet(document));
        return cssCompsRaw ? cssCompsRaw.items : [];
    }
    createSanitizedDocument(ast, filePath, version) {
        let cleanContentAst = this.cleanValuesInMediaQuery(ast);
        cleanContentAst = this.cleanStScopes(cleanContentAst);
        return vscode_languageserver_textdocument_1.TextDocument.create(vscode_uri_1.URI.file(filePath).toString(), 'stylable', version, cleanContentAst.toString());
    }
    // cleaning strategy: replace the "value(*)" syntax with spaces,
    // without touching the inner content (*)
    // by removing the function from the params,
    // the css-service will consider it a valid media query
    cleanValuesInMediaQuery(ast) {
        const mq = 'media';
        const valueMatch = 'value(';
        ast.walkAtRules(mq, (atRule) => {
            while (atRule.params.includes('value(')) {
                const currentValueIndex = atRule.params.indexOf(valueMatch);
                const closingParenthesisIndex = atRule.params.indexOf(')', currentValueIndex);
                atRule.params = atRule.params.replace(valueMatch, ' '.repeat(valueMatch.length));
                atRule.params =
                    atRule.params.substr(0, closingParenthesisIndex) +
                        ' ' +
                        atRule.params.substr(closingParenthesisIndex + 1);
            }
        });
        return ast;
    }
    // cleaning strategy: replace the "@st-scope" syntax with a media query,
    // if the @st-scope param was a class, replace the "." with a space as well.
    // this works because the css-service now considers this a valid media query where
    // completions and nesting behaviors are similar between the two
    cleanStScopes(ast) {
        const stScope = 'st-scope';
        const mq = 'media';
        ast.walkAtRules(stScope, (atRule) => {
            atRule.name = mq + ' '.repeat(stScope.length - mq.length);
            if (atRule.params.includes('.')) {
                atRule.params = atRule.params.replace('.', ' ');
            }
        });
        return ast;
    }
    getDiagnostics(document) {
        if (!document.uri.endsWith('.css')) {
            return [];
        }
        const stylesheet = this.inner.parseStylesheet(document);
        return this.inner
            .doValidation(document, stylesheet)
            .filter((diag) => {
            if (diag.code === 'emptyRules') {
                return false;
            }
            const atRuleName = readDocRange(document, diag.range);
            if (diag.code === 'unknownAtRules' &&
                (atRuleName === '@custom-selector' ||
                    atRuleName === '@st-scope' ||
                    atRuleName === '@st-import')) {
                return false;
            }
            if (diag.code === 'css-lcurlyexpected' &&
                readDocRange(document, vscode_languageserver_types_1.Range.create(vscode_languageserver_types_1.Position.create(diag.range.start.line, 0), diag.range.end)).startsWith('@custom-selector')) {
                return false;
            }
            if (diag.code === 'css-rparentexpected' || diag.code === 'css-identifierexpected') {
                const line = readDocRange(document, vscode_languageserver_types_1.Range.create(vscode_languageserver_types_1.Position.create(diag.range.start.line, 0), diag.range.end));
                const stateStart = findPseudoStateStart(line, diag.range.start.character);
                if (stateStart.index !== -1 && stateStart.openParens > 0) {
                    return false;
                }
            }
            if (diag.code === 'unknownProperties') {
                const prop = diag.message.match(/'(.*)'/)[1];
                const uri = vscode_uri_1.URI.parse(document.uri);
                // on windows, uri.fsPath replaces separators with '\'
                // this breaks posix paths in-memory when running on windows
                // take raw posix path instead
                const filePath = uri.scheme === 'file' &&
                    !uri.authority && // not UNC
                    uri.path.charCodeAt(2) !== 58 && // the colon in "c:"
                    path_1.default.isAbsolute(uri.path)
                    ? uri.path
                    : uri.fsPath;
                const src = this.fs.readFileSync(filePath, 'utf8');
                const meta = provider_1.createMeta(src, filePath).meta;
                if (meta && Object.keys(meta.mappedSymbols).some((ms) => ms === prop)) {
                    return false;
                }
            }
            return true;
        })
            .map((diag) => {
            diag.source = 'css';
            return diag;
        });
    }
    doHover(document, position, settings) {
        const stylesheet = this.inner.parseStylesheet(document);
        return this.inner.doHover(document, position, stylesheet, settings);
    }
    findReferences(document, position) {
        const stylesheet = this.inner.parseStylesheet(document);
        return this.inner.findReferences(document, position, stylesheet);
    }
    getColorPresentations(document, color, range) {
        const stylesheet = this.inner.parseStylesheet(document);
        return this.inner.getColorPresentations(document, stylesheet, color, range);
    }
    findColors(document) {
        const stylesheet = this.inner.parseStylesheet(document);
        return this.inner.findDocumentColors(document, stylesheet);
    }
    findColor(document) {
        const colors = this.findColors(document);
        return colors.length ? colors[0].color : null;
    }
}
exports.CssService = CssService;
//# sourceMappingURL=css-service.js.map